@@ -11,7 +11,7 @@ module Show_lock = struct
     in
     Console.print
     @@ List.map lock_dir_paths ~f:(fun lock_dir_path ->
-      let lock_dir = Lock_dir.read_disk lock_dir_path in
+      let lock_dir = Lock_dir.read_disk_exn lock_dir_path in
       Pp.concat
         ~sep:Pp.space
         [ Pp.hovbox
@@ -46,10 +46,11 @@ module Dependency_hash = struct
       >>| Package_name.Map.values
       >>| List.map ~f:Local_package.for_solver
     in
-    match
-      Local_package.(
-        For_solver.list_non_local_dependency_set local_packages |> Dependency_set.hash)
-    with
+    let hash =
+      Local_package.For_solver.non_local_dependencies local_packages
+      |> Local_package.Dependency_hash.of_dependency_formula
+    in
+    match hash with
     | None -> User_error.raise [ Pp.text "No non-local dependencies" ]
     | Some dependency_hash ->
       print_endline (Local_package.Dependency_hash.to_string dependency_hash)
@@ -120,7 +121,7 @@ module List_locked_dependencies = struct
     List.filter_map lock_dirs ~f:(fun lock_dir_path ->
       if Path.exists (Path.source lock_dir_path)
       then (
-        try Some (lock_dir_path, Lock_dir.read_disk lock_dir_path) with
+        try Some (lock_dir_path, Lock_dir.read_disk_exn lock_dir_path) with
         | User_error.E e ->
           User_warning.emit
             [ Pp.textf
