@@ -32,12 +32,16 @@ let interpret_destdir ~destdir path =
 ;;
 
 let get_dirs context ~prefix_from_command_line ~from_command_line =
+  let open Fiber.O in
   let module Roots = Install.Roots in
   let prefix_from_command_line = Option.map ~f:Path.of_string prefix_from_command_line in
-  let roots =
+  let+ roots =
     match prefix_from_command_line with
-    | Some prefix -> Roots.opam_from_prefix prefix |> Roots.map ~f:(fun s -> Some s)
-    | None -> Context.roots context
+    | None -> Memo.run (Context.roots context)
+    | Some prefix ->
+      Roots.opam_from_prefix prefix ~relative:Path.relative
+      |> Roots.map ~f:(fun s -> Some s)
+      |> Fiber.return
   in
   let roots = Roots.first_has_priority from_command_line roots in
   let must_be_defined name v =
@@ -79,9 +83,9 @@ module Workspace = struct
   let get () =
     let open Memo.O in
     Memo.run
-      (let+ conf = Dune_rules.Dune_load.load ()
+      (let+ packages = Dune_rules.Dune_load.packages ()
        and+ contexts = Context.DB.all () in
-       { packages = conf.packages; contexts })
+       { packages; contexts })
   ;;
 
   let package_install_file t ~findlib_toolchain pkg =
@@ -137,6 +141,10 @@ module Special_file = struct
   ;;
 end
 
+type copy_kind =
+  | Substitute (** Use [Artifact_substitution.copy_file]. Will scan all bytes. *)
+  | Special of Special_file.t (** Hooks to add version numbers, replace sections, etc *)
+
 type rmdir_mode =
   | Fail
   | Warn
@@ -147,7 +155,7 @@ module type File_operations = sig
     :  src:Path.t
     -> dst:Path.t
     -> executable:bool
-    -> special_file:Special_file.t option
+    -> kind:copy_kind
     -> package:Package.Name.t
     -> conf:Artifact_substitution.Conf.t
     -> unit Fiber.t
@@ -164,7 +172,7 @@ module File_ops_dry_run (Verbosity : sig
 
   let print_line fmt = print_line ~verbosity fmt
 
-  let copy_file ~src ~dst ~executable ~special_file:_ ~package:_ ~conf:_ =
+  let copy_file ~src ~dst ~executable ~kind:_ ~package:_ ~conf:_ =
     print_line
       "Copying %s to %s (executable: %b)"
       (Path.to_string_maybe_quoted src)
@@ -198,47 +206,37 @@ module File_ops_real (W : sig
   let print_line = print_line ~verbosity
   let get_vcs p = Source_tree.nearest_vcs p
 
-  type load_special_file_result =
-    { need_version : bool
-    ; callback : ?version:string -> Format.formatter -> unit
-    }
+  type copy_special_file_status =
+    | Done
+    | Use_plain_copy
+
+  let with_ppf oc ~f =
+    let ppf = Format.formatter_of_out_channel oc in
+    f ppf;
+    Format.pp_print_flush ppf ()
+  ;;
 
   let copy_special_file ~src ~package ~ic ~oc ~f =
     let open Fiber.O in
-    let plain_copy () =
-      seek_in ic 0;
-      Io.copy_channels ic oc;
-      Fiber.return ()
-    in
-    match f ic with
-    | Some { need_version; callback } ->
-      let* version =
-        if need_version
-        then
-          let* packages =
-            match Package.Name.Map.find workspace.packages package with
-            | None -> Fiber.return None
-            | Some package -> Memo.run (get_vcs (Package.dir package))
-          in
-          match packages with
-          | None -> Fiber.return None
-          | Some vcs -> Memo.run (Vcs.describe vcs)
-        else Fiber.return None
+    let get_version () =
+      let* packages =
+        match Package.Name.Map.find workspace.packages package with
+        | None -> Fiber.return None
+        | Some package -> Memo.run (get_vcs (Package.dir package))
       in
-      let ppf = Format.formatter_of_out_channel oc in
-      callback ppf ?version;
-      Format.pp_print_flush ppf ();
-      Fiber.return ()
-    | None -> plain_copy ()
-    (* XXX should we really be catching everything here? *)
-    | exception _ ->
+      match packages with
+      | None -> Fiber.return None
+      | Some vcs -> Memo.run (Vcs.describe vcs)
+    in
+    try f ~get_version ic ~src oc with
+    | _ (* XXX should we really be catching everything here? *) ->
       User_warning.emit
         ~loc:(Loc.in_file src)
         [ Pp.text "Failed to parse file, not adding version and locations information." ];
-      plain_copy ()
+      Fiber.return Use_plain_copy
   ;;
 
-  let process_meta ic =
+  let process_meta ~get_version ic ~src:_ oc =
     let module Meta = Dune_findlib.Findlib.Meta in
     let lb = Lexing.from_channel ic in
     let meta : Meta.t = { name = None; entries = Meta.parse_entries lb } in
@@ -252,134 +250,108 @@ module File_ops_real (W : sig
       | Exit -> true
     in
     if not need_more_versions
-    then None
-    else (
-      let callback ?version ppf =
+    then Fiber.return Use_plain_copy
+    else
+      let open Fiber.O in
+      let+ version = get_version () in
+      with_ppf oc ~f:(fun ppf ->
         let meta = Meta.add_versions meta ~get_version:(fun _ -> version) in
-        Pp.to_fmt ppf (Meta.pp meta.entries)
-      in
-      Some { need_version = true; callback })
+        Pp.to_fmt ppf (Meta.pp meta.entries));
+      Done
   ;;
 
-  let replace_sites ~(get_location : Section.t -> Package.Name.t -> Stdune.Path.t) dp =
-    match
-      List.find_map dp ~f:(function
-        | Dune_lang.List [ Atom (A "name"); Atom (A name) ] -> Some name
-        | _ -> None)
-    with
-    | None -> dp
-    | Some name ->
-      List.map dp ~f:(function
-        | Dune_lang.List ((Atom (A "sections") as sexp_sections) :: sections) ->
-          let sections =
-            List.map sections ~f:(function
-              | Dune_lang.List [ (Atom (A section) as section_sexp); _ ] ->
-                let path =
-                  get_location
-                    (Option.value_exn (Section.of_string section))
-                    (Package.Name.of_string name)
-                in
-                let open Dune_lang.Encoder in
-                pair sexp string (section_sexp, Path.to_absolute_filename path)
-              | _ -> assert false)
-          in
-          Dune_lang.List (sexp_sections :: sections)
-        | x -> x)
-  ;;
-
-  let process_dune_package ~get_location ic =
+  let process_dune_package ~get_version ~get_location ic ~src oc =
     let lb = Lexing.from_channel ic in
-    let dp =
-      Dune_lang.Parser.parse ~mode:Many lb |> List.map ~f:Dune_lang.Ast.remove_locs
-    in
-    (* replace sites with external path in the file *)
-    let dp = replace_sites ~get_location dp in
-    (* replace version if needed in the file *)
-    let need_version =
-      not
-        (List.exists dp ~f:(function
-          | Dune_lang.List (Atom (A "version") :: _)
-          | Dune_lang.List [ Atom (A "use_meta"); Atom (A "true") ]
-          | Dune_lang.List [ Atom (A "use_meta") ] -> true
-          | _ -> false))
-    in
-    let callback ?version ppf =
-      let dp =
-        match version with
-        | Some version ->
-          let version =
-            Dune_lang.List
-              [ Dune_lang.atom "version"; Dune_lang.atom_or_quoted_string version ]
-          in
-          (match dp with
-           | lang :: name :: rest -> lang :: name :: version :: rest
-           | [ lang ] -> [ lang; version ]
-           | [] -> [ version ])
-        | _ -> dp
+    let dune_version = Dune_lang.Syntax.greatest_supported_version_exn Stanza.syntax in
+    match Dune_package.Or_meta.parse src lb |> User_error.ok_exn with
+    | Use_meta ->
+      with_ppf oc ~f:(Dune_package.Or_meta.pp_use_meta ~dune_version);
+      Fiber.return Done
+    | Dune_package dp ->
+      let open Fiber.O in
+      (* replace sites with external path in the file *)
+      let dp, replace_info = Dune_package.replace_site_sections ~get_location dp in
+      (* replace version if needed in the file *)
+      let need_version = Option.is_none dp.version in
+      let+ dp =
+        if need_version
+        then
+          let+ version_opt = get_version () in
+          match version_opt with
+          | Some version -> { dp with version = Some (Package_version.of_string version) }
+          | None -> dp
+        else Fiber.return dp
       in
-      Format.pp_open_vbox ppf 0;
-      List.iter dp ~f:(fun x ->
-        Dune_lang.Deprecated.pp ppf x;
-        Format.pp_print_cut ppf ());
-      Format.pp_close_box ppf ()
-    in
-    Some { need_version; callback }
+      with_ppf oc ~f:(fun ppf ->
+        (* CR-emillon: we should write absolute paths only if necessary *)
+        Dune_package.Or_meta.pp
+          ~dune_version
+          ppf
+          (Dune_package dp)
+          ~encoding:(Absolute replace_info));
+      Done
   ;;
 
   let copy_file
     ~src
     ~dst
     ~executable
-    ~special_file
+    ~kind
     ~package
     ~(conf : Artifact_substitution.Conf.t)
     =
     let chmod = if executable then fun _ -> 0o755 else fun _ -> 0o644 in
-    match (special_file : Special_file.t option) with
-    | None -> Artifact_substitution.copy_file ~conf ~executable ~src ~dst ~chmod ()
-    | Some sf ->
+    let plain_copy () = Io.copy_file ~chmod ~src ~dst () in
+    match kind with
+    | Substitute -> Artifact_substitution.copy_file ~conf ~src ~dst ~chmod ()
+    | Special sf ->
+      let open Fiber.O in
       let ic, oc = Io.setup_copy ~chmod ~src ~dst () in
-      Fiber.finalize
-        ~finally:(fun () ->
-          Io.close_both (ic, oc);
-          Fiber.return ())
-        (fun () ->
-          let f =
-            match sf with
-            | META -> process_meta
-            | Dune_package ->
-              process_dune_package
-                ~get_location:(Artifact_substitution.Conf.get_location conf)
-          in
-          copy_special_file ~src ~package ~ic ~oc ~f)
+      let+ status =
+        Fiber.finalize
+          ~finally:(fun () ->
+            Io.close_both (ic, oc);
+            Fiber.return ())
+          (fun () ->
+            let f =
+              match sf with
+              | META -> process_meta
+              | Dune_package ->
+                process_dune_package
+                  ~get_location:(Artifact_substitution.Conf.get_location conf)
+            in
+            copy_special_file ~src ~package ~ic ~oc ~f)
+      in
+      (match status with
+       | Done -> ()
+       | Use_plain_copy -> plain_copy ())
   ;;
 
   let remove_file_if_exists dst =
     if Path.exists dst
     then (
       print_line "Deleting %s" (Path.to_string_maybe_quoted dst);
-      print_unix_error (fun () -> Path.unlink dst))
+      print_unix_error (fun () -> Path.unlink_exn dst))
   ;;
 
   let remove_dir_if_exists ~if_non_empty dir =
-    if Path.exists dir
-    then (
-      match Path.readdir_unsorted dir with
-      | Ok [] ->
-        print_line "Deleting empty directory %s" (Path.to_string_maybe_quoted dir);
-        print_unix_error (fun () -> Path.rmdir dir)
-      | Error (e, _, _) ->
-        User_message.prerr (User_error.make [ Pp.text (Unix.error_message e) ])
-      | _ ->
-        let dir = Path.to_string_maybe_quoted dir in
-        (match if_non_empty with
-         | Warn ->
-           User_message.prerr
-             (User_error.make
-                [ Pp.textf "Directory %s is not empty, cannot delete (ignoring)." dir ])
-         | Fail ->
-           User_error.raise
-             [ Pp.textf "Please delete non-empty directory %s manually." dir ]))
+    match Path.readdir_unsorted dir with
+    | Error (Unix.ENOENT, _, _) -> ()
+    | Ok [] ->
+      print_line "Deleting empty directory %s" (Path.to_string_maybe_quoted dir);
+      print_unix_error (fun () -> Path.rmdir dir)
+    | Error (e, _, _) ->
+      User_message.prerr (User_error.make [ Pp.text (Unix.error_message e) ])
+    | _ ->
+      let dir = Path.to_string_maybe_quoted dir in
+      (match if_non_empty with
+       | Warn ->
+         User_message.prerr
+           (User_error.make
+              [ Pp.textf "Directory %s is not empty, cannot delete (ignoring)." dir ])
+       | Fail ->
+         User_error.raise
+           [ Pp.textf "Please delete non-empty directory %s manually." dir ])
   ;;
 
   let mkdir_p p =
@@ -459,7 +431,265 @@ let cmd_what = function
   | Uninstall -> "uninstall"
 ;;
 
-let install_uninstall ~what =
+let install_entry
+  ~ops
+  ~conf
+  ~package
+  ~dir
+  ~create_install_files
+  (entry : Path.t Install.Entry.t)
+  ~dst
+  ~verbosity
+  =
+  let module Ops = (val ops : File_operations) in
+  let open Fiber.O in
+  let special_file = Special_file.of_entry entry in
+  (match special_file with
+   | _ when not create_install_files -> Fiber.return true
+   | Some Special_file.META | Some Special_file.Dune_package -> Fiber.return true
+   | None ->
+     Artifact_substitution.test_file ~src:entry.src ()
+     >>| (function
+      | Some_substitution -> true
+      | No_substitution -> false))
+  >>= function
+  | false -> Fiber.return entry
+  | true ->
+    let+ () =
+      (match Path.is_directory dst with
+       | true -> Ops.remove_dir_if_exists ~if_non_empty:Fail dst
+       | false -> Ops.remove_file_if_exists dst);
+      print_line
+        ~verbosity
+        "%s %s"
+        (if create_install_files then "Copying to" else "Installing")
+        (Path.to_string_maybe_quoted dst);
+      Ops.mkdir_p dir;
+      let executable = Section.should_set_executable_bit entry.section in
+      let kind =
+        match special_file with
+        | Some special -> Special special
+        | None ->
+          (* CR-emillon: for most cases we could use a fast copy here, but some
+             kinds of files do need artifact substitution(at least
+             executable files and artifacts built from generated sites
+             modules), but it's too late to know without reading the file. *)
+          Substitute
+      in
+      Ops.copy_file ~src:entry.src ~dst ~executable ~kind ~package ~conf
+    in
+    Install.Entry.set_src entry dst
+;;
+
+let run
+  what
+  context
+  common
+  pkgs
+  sections
+  (config : Dune_config.t)
+  ~dry_run
+  ~destdir
+  ~relocatable
+  ~create_install_files
+  ~prefix_from_command_line
+  ~(from_command_line : _ Install.Roots.t)
+  =
+  let open Fiber.O in
+  let* workspace = Workspace.get () in
+  let contexts =
+    match context with
+    | None ->
+      (match Common.x common with
+       | Some findlib_toolchain ->
+         let contexts =
+           List.filter workspace.contexts ~f:(fun (ctx : Context.t) ->
+             match Context.findlib_toolchain ctx with
+             | None -> false
+             | Some ctx_findlib_toolchain ->
+               Dune_engine.Context_name.equal ctx_findlib_toolchain findlib_toolchain)
+         in
+         contexts
+       | None -> workspace.contexts)
+    | Some name ->
+      (match
+         List.find workspace.contexts ~f:(fun c ->
+           Dune_engine.Context_name.equal (Context.name c) name)
+       with
+       | Some ctx -> [ ctx ]
+       | None ->
+         User_error.raise
+           [ Pp.textf "Context %S not found!" (Dune_engine.Context_name.to_string name) ])
+  in
+  let* pkgs =
+    match pkgs with
+    | _ :: _ -> Fiber.return pkgs
+    | [] ->
+      Package.Name.Map.values workspace.packages
+      |> Fiber.parallel_map ~f:(fun pkg ->
+        package_is_vendored pkg
+        >>| function
+        | true -> None
+        | false -> Some (Package.name pkg))
+      >>| List.filter_opt
+  in
+  let install_files, missing_install_files =
+    List.concat_map pkgs ~f:(fun pkg ->
+      List.map contexts ~f:(fun (ctx : Context.t) ->
+        let fn =
+          let fn =
+            resolve_package_install
+              workspace
+              ~findlib_toolchain:(Context.findlib_toolchain ctx)
+              pkg
+          in
+          Path.append_source (Path.build (Context.build_dir ctx)) fn
+        in
+        if Path.exists fn then Left (ctx, (pkg, fn)) else Right fn))
+    |> List.partition_map ~f:Fun.id
+  in
+  if missing_install_files <> []
+  then
+    User_error.raise
+      [ Pp.textf "The following <package>.install are missing:"
+      ; Pp.enumerate missing_install_files ~f:(fun p -> Pp.text (Path.to_string p))
+      ]
+      ~hints:
+        [ Pp.concat
+            ~sep:Pp.space
+            [ Pp.text "try running"
+            ; User_message.command "dune build [-p <pkg>] @install"
+            ]
+          |> Pp.hovbox
+        ];
+  (match contexts, prefix_from_command_line, from_command_line.lib_root with
+   | _ :: _ :: _, Some _, _ | _ :: _ :: _, _, Some _ ->
+     User_error.raise
+       [ Pp.concat
+           ~sep:Pp.space
+           [ Pp.text "Cannot specify"
+           ; User_message.command "--prefix"
+           ; Pp.text "or"
+           ; User_message.command "--libdir"
+           ; Pp.text "when installing into multiple contexts!"
+           ]
+       ]
+   | _ -> ());
+  let install_files_by_context =
+    let module CMap = Map.Make (Context) in
+    CMap.of_list_multi install_files
+    |> CMap.to_list_map ~f:(fun context install_files ->
+      let entries_per_package =
+        List.map install_files ~f:(fun (package, install_file) ->
+          let entries =
+            Install.Entry.load_install_file install_file Path.of_local
+            |> List.filter ~f:(fun (entry : Path.t Install.Entry.t) ->
+              Sections.should_install sections entry.section)
+          in
+          match
+            List.filter_map entries ~f:(fun entry ->
+              (* CR rgrinberg: this is ignoring optional entries *)
+              Option.some_if (not (Path.exists entry.src)) entry.src)
+          with
+          | [] -> package, entries
+          | missing_files ->
+            User_error.raise
+              [ Pp.textf
+                  "The following files which are listed in %s cannot be installed \
+                   because they do not exist:"
+                  (Path.to_string_maybe_quoted install_file)
+              ; Pp.enumerate missing_files ~f:(fun p ->
+                  Pp.verbatim (Path.to_string_maybe_quoted p))
+              ])
+      in
+      context, entries_per_package)
+  in
+  let destdir =
+    Option.map
+      ~f:Path.of_string
+      (if create_install_files
+       then
+         (* CR-rgrinberg: why are we silently ignoring an argument instead
+            of erroring given that they mutually exclusive? *)
+         Some (Option.value ~default:"_destdir" destdir)
+       else destdir)
+  in
+  let relocatable =
+    if relocatable
+    then (
+      match prefix_from_command_line with
+      | Some dir -> Some (Path.of_string dir)
+      | None ->
+        User_error.raise
+          [ Pp.concat
+              ~sep:Pp.space
+              [ Pp.text "Option"
+              ; User_message.command "--prefix"
+              ; Pp.text "is needed with"
+              ; User_message.command "--relocation"
+              ]
+            |> Pp.hovbox
+          ])
+    else None
+  in
+  let verbosity =
+    match config.display with
+    | Simple display -> display.verbosity
+    | Tui -> Quiet
+  in
+  let open Fiber.O in
+  let (module Ops) = file_operations ~verbosity ~dry_run ~workspace in
+  let files_deleted_in = ref Path.Set.empty in
+  let+ () =
+    Fiber.sequential_iter
+      install_files_by_context
+      ~f:(fun (context, entries_per_package) ->
+        let* roots = get_dirs context ~prefix_from_command_line ~from_command_line in
+        let conf = Artifact_substitution.Conf.of_install ~relocatable ~roots ~context in
+        Fiber.sequential_iter entries_per_package ~f:(fun (package, entries) ->
+          let+ entries =
+            (* CR rgrinberg: why don't we install things concurrently? *)
+            Fiber.sequential_map entries ~f:(fun entry ->
+              let dst =
+                let paths = Install.Paths.make ~relative:Path.relative ~package ~roots in
+                Install.Entry.relative_installed_path entry ~paths
+                |> interpret_destdir ~destdir
+              in
+              let dir = Path.parent_exn dst in
+              match what with
+              | Uninstall ->
+                Ops.remove_file_if_exists dst;
+                files_deleted_in := Path.Set.add !files_deleted_in dir;
+                Fiber.return entry
+              | Install ->
+                install_entry
+                  ~ops:(module Ops)
+                  ~conf
+                  ~package
+                  ~dir
+                  ~create_install_files
+                  ~dst
+                  ~verbosity
+                  entry)
+          in
+          if create_install_files
+          then (
+            let fn =
+              resolve_package_install
+                workspace
+                ~findlib_toolchain:(Context.findlib_toolchain context)
+                package
+            in
+            Install.Entry.gen_install_file entries |> Io.write_file (Path.source fn))))
+  in
+  Path.Set.to_list !files_deleted_in
+  (* This [List.rev] is to ensure we process children directories before
+     their parents *)
+  |> List.rev
+  |> List.iter ~f:(Ops.remove_dir_if_exists ~if_non_empty:Warn)
+;;
+
+let make ~what =
   let doc = Format.asprintf "%a packages defined in the workspace." pp_what what in
   let name_ = Arg.info [] ~docv:"PACKAGE" in
   let absolute_path =
@@ -587,157 +817,8 @@ let install_uninstall ~what =
     let builder = Common.Builder.disable_log_file builder in
     let common, config = Common.init builder in
     Scheduler.go ~common ~config (fun () ->
-      let open Fiber.O in
-      let* workspace = Workspace.get () in
-      let contexts =
-        match context with
-        | None ->
-          (match Common.x common with
-           | Some findlib_toolchain ->
-             let contexts =
-               List.filter workspace.contexts ~f:(fun (ctx : Context.t) ->
-                 match Context.findlib_toolchain ctx with
-                 | None -> false
-                 | Some ctx_findlib_toolchain ->
-                   Dune_engine.Context_name.equal ctx_findlib_toolchain findlib_toolchain)
-             in
-             contexts
-           | None -> workspace.contexts)
-        | Some name ->
-          (match
-             List.find workspace.contexts ~f:(fun c ->
-               Dune_engine.Context_name.equal (Context.name c) name)
-           with
-           | Some ctx -> [ ctx ]
-           | None ->
-             User_error.raise
-               [ Pp.textf
-                   "Context %S not found!"
-                   (Dune_engine.Context_name.to_string name)
-               ])
-      in
-      let* pkgs =
-        match pkgs with
-        | [] ->
-          Package.Name.Map.values workspace.packages
-          |> Fiber.parallel_map ~f:(fun pkg ->
-            package_is_vendored pkg
-            >>| function
-            | true -> None
-            | false -> Some (Package.name pkg))
-          >>| List.filter_opt
-        | l -> Fiber.return l
-      in
-      let install_files, missing_install_files =
-        List.concat_map pkgs ~f:(fun pkg ->
-          List.map contexts ~f:(fun (ctx : Context.t) ->
-            let fn =
-              let fn =
-                resolve_package_install
-                  workspace
-                  ~findlib_toolchain:(Context.findlib_toolchain ctx)
-                  pkg
-              in
-              Path.append_source (Path.build (Context.build_dir ctx)) fn
-            in
-            if Path.exists fn then Left (ctx, (pkg, fn)) else Right fn))
-        |> List.partition_map ~f:Fun.id
-      in
-      if missing_install_files <> []
-      then
-        User_error.raise
-          [ Pp.textf "The following <package>.install are missing:"
-          ; Pp.enumerate missing_install_files ~f:(fun p -> Pp.text (Path.to_string p))
-          ]
-          ~hints:
-            [ Pp.concat
-                ~sep:Pp.space
-                [ Pp.text "try running"
-                ; User_message.command "dune build [-p <pkg>] @install"
-                ]
-              |> Pp.hovbox
-            ];
-      (match contexts, prefix_from_command_line, libdir_from_command_line with
-       | _ :: _ :: _, Some _, _ | _ :: _ :: _, _, Some _ ->
-         User_error.raise
-           [ Pp.concat
-               ~sep:Pp.space
-               [ Pp.text "Cannot specify"
-               ; User_message.command "--prefix"
-               ; Pp.text "or"
-               ; User_message.command "--libdir"
-               ; Pp.text "when installing into multiple contexts!"
-               ]
-           ]
-       | _ -> ());
-      let install_files_by_context =
-        let module CMap = Map.Make (Context) in
-        CMap.of_list_multi install_files
-        |> CMap.to_list_map ~f:(fun context install_files ->
-          let entries_per_package =
-            List.map install_files ~f:(fun (package, install_file) ->
-              let entries =
-                Install.Entry.load_install_file install_file (fun local ->
-                  Path.append_local (Path.source Path.Source.root) local)
-              in
-              let entries =
-                List.filter entries ~f:(fun (entry : Path.t Install.Entry.t) ->
-                  Sections.should_install sections entry.section)
-              in
-              match
-                List.filter_map entries ~f:(fun entry ->
-                  (* CR rgrinberg: this is ignoring optional entries *)
-                  Option.some_if (not (Path.exists entry.src)) entry.src)
-              with
-              | [] -> package, entries
-              | missing_files ->
-                User_error.raise
-                  [ Pp.textf
-                      "The following files which are listed in %s cannot be installed \
-                       because they do not exist:"
-                      (Path.to_string_maybe_quoted install_file)
-                  ; Pp.enumerate missing_files ~f:(fun p ->
-                      Pp.verbatim (Path.to_string_maybe_quoted p))
-                  ])
-          in
-          context, entries_per_package)
-      in
-      let destdir =
-        Option.map
-          ~f:Path.of_string
-          (if create_install_files
-           then Some (Option.value ~default:"_destdir" destdir)
-           else destdir)
-      in
-      let relocatable =
-        if relocatable
-        then (
-          match prefix_from_command_line with
-          | Some dir -> Some (Path.of_string dir)
-          | None ->
-            User_error.raise
-              [ Pp.concat
-                  ~sep:Pp.space
-                  [ Pp.text "Option"
-                  ; User_message.command "--prefix"
-                  ; Pp.text "is needed with"
-                  ; User_message.command "--relocation"
-                  ]
-                |> Pp.hovbox
-              ])
-        else None
-      in
-      let verbosity =
-        match config.display with
-        | Simple display -> display.verbosity
-        | Tui -> Quiet
-      in
-      let open Fiber.O in
-      let (module Ops) = file_operations ~verbosity ~dry_run ~workspace in
-      let files_deleted_in = ref Path.Set.empty in
       let from_command_line =
-        let open Install.Roots in
-        { lib_root = libdir_from_command_line
+        { Install.Roots.lib_root = libdir_from_command_line
         ; etc_root = etcdir_from_command_line
         ; doc_root = docdir_from_command_line
         ; man = mandir_from_command_line
@@ -746,90 +827,22 @@ let install_uninstall ~what =
         ; libexec_root = libexecdir_from_command_line
         ; share_root = datadir_from_command_line
         }
-        |> map ~f:(Option.map ~f:Path.of_string)
-        |> complete
+        |> Install.Roots.map ~f:(Option.map ~f:Path.of_string)
+        |> Install.Roots.complete
       in
-      let+ () =
-        Fiber.sequential_iter
-          install_files_by_context
-          ~f:(fun (context, entries_per_package) ->
-            let roots = get_dirs context ~prefix_from_command_line ~from_command_line in
-            let conf =
-              Artifact_substitution.Conf.of_install ~relocatable ~roots ~context
-            in
-            Fiber.sequential_iter entries_per_package ~f:(fun (package, entries) ->
-              let paths = Install.Paths.make ~package ~roots in
-              let+ entries =
-                (* CR rgrinberg: why don't we install things concurrently? *)
-                Fiber.sequential_map entries ~f:(fun entry ->
-                  let special_file = Special_file.of_entry entry in
-                  let dst =
-                    Install.Entry.relative_installed_path entry ~paths
-                    |> interpret_destdir ~destdir
-                  in
-                  let dir = Path.parent_exn dst in
-                  match what with
-                  | Install ->
-                    let* copy =
-                      match special_file with
-                      | _ when not create_install_files -> Fiber.return true
-                      | None ->
-                        let open Artifact_substitution in
-                        let+ status = test_file ~src:entry.src () in
-                        (match status with
-                         | Some_substitution -> true
-                         | No_substitution -> false)
-                      | Some Special_file.META | Some Special_file.Dune_package ->
-                        Fiber.return true
-                    in
-                    let msg =
-                      if create_install_files then "Copying to" else "Installing"
-                    in
-                    if copy
-                    then
-                      let* () =
-                        (match Path.is_directory dst with
-                         | true -> Ops.remove_dir_if_exists ~if_non_empty:Fail dst
-                         | false -> Ops.remove_file_if_exists dst);
-                        print_line
-                          ~verbosity
-                          "%s %s"
-                          msg
-                          (Path.to_string_maybe_quoted dst);
-                        Ops.mkdir_p dir;
-                        let executable =
-                          Section.should_set_executable_bit entry.section
-                        in
-                        Ops.copy_file
-                          ~src:entry.src
-                          ~dst
-                          ~executable
-                          ~special_file
-                          ~package
-                          ~conf
-                      in
-                      Fiber.return (Install.Entry.set_src entry dst)
-                    else Fiber.return entry
-                  | Uninstall ->
-                    Ops.remove_file_if_exists dst;
-                    files_deleted_in := Path.Set.add !files_deleted_in dir;
-                    Fiber.return entry)
-              in
-              if create_install_files
-              then (
-                let fn =
-                  resolve_package_install
-                    workspace
-                    ~findlib_toolchain:(Context.findlib_toolchain context)
-                    package
-                in
-                Io.write_file (Path.source fn) (Install.Entry.gen_install_file entries))))
-      in
-      Path.Set.to_list !files_deleted_in
-      (* This [List.rev] is to ensure we process children directories before
-         their parents *)
-      |> List.rev
-      |> List.iter ~f:(Ops.remove_dir_if_exists ~if_non_empty:Warn))
+      run
+        what
+        context
+        common
+        pkgs
+        sections
+        config
+        ~dry_run
+        ~destdir
+        ~relocatable
+        ~create_install_files
+        ~prefix_from_command_line
+        ~from_command_line)
   in
   Cmd.v
     (Cmd.info
@@ -839,5 +852,5 @@ let install_uninstall ~what =
     term
 ;;
 
-let install = install_uninstall ~what:Install
-let uninstall = install_uninstall ~what:Uninstall
+let install = make ~what:Install
+let uninstall = make ~what:Uninstall
