@@ -9,8 +9,6 @@ type t = private
   ; related : User_message.t list
   }
 
-val annot : t User_message.Annots.Key.t
-
+val annot : t list User_message.Annots.Key.t
 val make : main:User_message.t -> related:User_message.t list -> t
-
-val parse_output : dir:Path.t -> string -> t option
+val parse_output : dir:Path.t -> string -> t list
