@@ -9,22 +9,13 @@ module Pkg : sig
   val default_exn : loc:Loc.t -> Dune_project.t -> string -> Package.t
 end
 
-(** [preprocess] and [preprocessor_deps] fields *)
-val preprocess_fields
-  : (Preprocess.Without_instrumentation.t Preprocess.Per_module.t * Dep_conf.t list)
-      fields_parser
-
-(** [instrumentation] multi field *)
-val instrumentation
-  : (Loc.t * (((Loc.t * Lib_name.t) * String_with_vars.t list) * Dep_conf.t list) list)
-      fields_parser
-
 module Modules_settings : sig
   type t =
     { root_module : (Loc.t * Module_name.t) option
-    ; modules_without_implementation : Ordered_set_lang.t
-    ; modules : Ordered_set_lang.t
+    ; modules_without_implementation : Ordered_set_lang.Unexpanded.t
+    ; modules : Ordered_set_lang.Unexpanded.t
     }
 
+  val since_expanded : Syntax.Version.t
   val decode : t Dune_lang.Decoder.fields_parser
 end
