@@ -308,7 +308,7 @@ let%expect_test _ =
   [%expect
     {|
       (Some [ (Some "lazy_memo", "foo"); (Some "id", "lazy: foo") ],
-      Some [ (Some "lazy_memo", "foo"); (Some "id", "lazy: foo") ])
+       Some [ (Some "lazy_memo", "foo"); (Some "id", "lazy: foo") ])
     |}]
 ;;
 
@@ -331,7 +331,7 @@ let%expect_test _ =
   [%expect
     {|
     (Some [ (Some "lazy_memo", "foo"); (None, ()) ],
-    Some [ (Some "lazy_memo", "foo"); (None, ()) ])
+     Some [ (Some "lazy_memo", "foo"); (None, ()) ])
   |}]
 ;;
 
@@ -486,11 +486,11 @@ let%expect_test "Memo.Poly" =
   let (s2 : string Function.input) = S (Type_eq.Id.create (), "hi again") in
   let run_int i =
     let res = run (M.eval i) in
-    Dyn.to_string (Dyn.List (List.map res ~f:Int.to_dyn))
+    Dyn.to_string (Dyn.list Dyn.int res)
   in
   let run_string s =
     let res = run (M.eval s) in
-    Dyn.to_string (Dyn.List (List.map res ~f:String.to_dyn))
+    Dyn.to_string (Dyn.list Dyn.string res)
   in
   printf "----- First-time calls -----\n";
   printf "%d -> %s\n" (Function.get i1) (run_int i1);
@@ -761,7 +761,7 @@ let%expect_test "dynamic cycles with non-uniform cutoff structure" =
     printf "Started evaluating %s\n" which;
     let* base = Memo.exec base () in
     match base with
-    | input when input = 2 ->
+    | 2 as input ->
       let summit = Fdecl.get summit_fdecl in
       printf "Cycling to summit from %s...\n" which;
       let+ result = Memo.exec summit input in
